-module(event_mgr).

-behaviour(gen_server).

-export([start_link/1, init/1, handle_cast/2, handle_call/3, handle_info/2]).
-export([update_config/2]).

-record(event_config, {database_file = none, skills_fields = [], practice_fields = [], divisions = []}).

-record(division_config, {fields = [], teams = [], teams_hash = "", practice_matches = [], practice_rounds = 0, practice_seed = 0, qualification_matches = [], qualification_rounds = [], qualification_seed = 0, elimination_alliances = 0}).

-record(event_state, {owner, config = #event_config{}, fields = [], database = none, supervisor}).

start_link(Args) ->
  gen_server:start_link({local, ?MODULE}, ?MODULE, Args, []).

init([Owner, none]) ->
  {ok, Supervisor} = event_sup:start_link(),
  {ok, #event_state{owner = Owner, supervisor = Supervisor}};
init([Owner, ConfigJSON]) ->
  {ok, Supervisor} = event_sup:start_link(),
  {ok, parse_event_config(ConfigJSON, #event_state{owner = Owner, supervisor = Supervisor})}.

get_matches_from_db(Database, Division, Round, Seed, Size, TeamsHash) ->
  if Size == 0 -> [];
     Size > 0 -> SQL = io_lib:format("SELECT number, blue_1, blue_2, red_1, red_2 FROM matches WHERE division = '~p' AND round = '~p' AND size = '~p' AND seed = '~p' AND teams = '~s';", [Division, Round, Size, Seed, TeamsHash]),
                 [{columns, Columns}, {rows, Rows}] = sqlite3:sql_exec(Database, SQL),
                 []
  end.

parse_division([], _, Parsed, Fields, _) -> {Parsed, Fields};
parse_division([Division | Divisions], Database, Parsed, Fields, N) ->
  TeamsBinary = maps:get(<<"teams">>, Division),
  TeamsList = [binary_to_list(X) || X <- TeamsBinary],
  TeamsHash = schedule:hash_teams_list(TeamsList),

  PracticeRounds = maps:get(<<"practice_rounds">>, Division),
  PracticeSeed = maps:get(<<"practice_seed">>, Division),
  PracticeMatches = get_matches_from_db(Database, N, practice, PracticeSeed, PracticeRounds, TeamsHash),

  QualificationRounds = maps:get(<<"qualification_rounds">>, Division),
  QualificationSeed = maps:get(<<"qualification_seed">>, Division),
  QualificationMatches = get_matches_from_db(Database, N, qualification, QualificationSeed, PracticeRounds, TeamsHash),

  DivConfig = #division_config{
                fields = [binary_to_list(X) || X <- maps:get(<<"fields">>, Division)],
                elimination_alliances = maps:get(<<"elimination_alliances">>, Division),
                practice_rounds = PracticeRounds,
                practice_matches = PracticeMatches,
                practice_seed = PracticeSeed,
                qualification_rounds = QualificationRounds,
                qualification_matches = QualificationMatches,
                qualification_seed = QualificationSeed,
                teams = TeamsList,
                teams_hash = TeamsHash
                },
  parse_division(Divisions, Database, [DivConfig | Parsed], lists:append([DivConfig#division_config.fields, Fields]), N+1).

parse_divisions(Divisions, Database) ->
  parse_division(Divisions, Database, [], [], 0).


parse_event_config(ConfigJSON, State) ->
  Config = jsone:decode(list_to_binary(ConfigJSON)),
  CurrentFile = State#event_state.config#event_config.database_file,
  NewFile = binary_to_list(maps:get(<<"database">>, Config)),
  {ok, Database} = if CurrentFile == NewFile -> {ok, State#event_state.database};
                      CurrentFile == none -> sqlite3:open(anonymous, [{file, NewFile}]);
                      true -> sqlite3:close(State#event_state.database), sqlite3:open(anonymous, [{file, NewFile}])
  end,
  {Divisions, DivisionFields} = parse_divisions(maps:get(<<"divisions">>, Config), Database),
  PracticeFields = [binary_to_list(X) || X <- maps:get(<<"testing_fields">>, Config)],
  SkillsFields = [binary_to_list(X) || X <- maps:get(<<"skills_fields">>, Config)],
  EliminationFields = [binary_to_list(X) || X <- maps:get(<<"elimination_fields">>, Config)],
  Fields = sets:to_list(sets:from_list(lists:append([DivisionFields, SkillsFields, PracticeFields, EliminationFields]))),

  FieldStates = add_fields(Fields, State#event_state.fields),
  
  State#event_state{config = #event_config{
                                database_file = binary_to_list(maps:get(<<"database">>, Config)),
                                skills_fields = SkillsFields,
                                practice_fields = PracticeFields,
                                divisions = Divisions
                               }, fields = FieldStates, database = Database}.

update_config(Server, ConfigJSON) ->
  gen_server:call(Server, {event_config, ConfigJSON}).

get_field_state(_, []) -> free;
get_field_state(Name, [{FieldName, FieldState} | FieldStates]) ->
  if Name == FieldName -> FieldState;
     true -> get_field_state(Name, FieldStates)
  end.

add_fields(Fields, FieldStates) ->
  [{Name, get_field_state(Name, FieldStates)} || Name <- Fields].


update_field({Name, free}, State) ->
  % TODO: Calculate which event should be running on this field
  % Priority List Logic:
  %  1) if testing field, assign testing event
  %  2) if skills field, assign skills event
  %  3) if division field
  %    0) in general, assign match if: (state != done && (num % num_fields) == field_index)
  %    1) Get practice match from DB
  %    2) Get qualification match from DB
  %    2) Get lists of elimination matches from DB, assign(and generate if necessary) first where:
  FieldState = free,
  {{Name, FieldState}, State};
update_field({Name, FieldState}, State) -> {{Name, FieldState}, State}.

update_events(State) ->
  % TODO: calculate which events should be running
  % TODO: stop events on fields that no longer exist
  % TODO: stop events on fields that are running the wrong event
  % TODO: start every event that should be running given the config
  {NewFields, NewState} = lists:mapfoldl(fun update_field/2, State, State#event_state.fields),
  NewState#event_state{fields = NewFields}.

handle_call({event_config, ConfigJSON}, _, State) ->
  NewState = parse_event_config(ConfigJSON, State),
  State#event_state.owner ! {new_event_config, NewState#event_state.config},
  {reply, ok, NewState};

handle_call(_, _, State) ->
  {noreply, update_events(State)}.

handle_cast(_, State) ->
  {noreply, update_events(State)}.

handle_info(_, State) ->
  {noreply, update_events(State)}.
