-module(schedule).

-behaviour(gen_server).

-export([start_link/0, init/1, handle_cast/2, handle_call/3]).

-export([init_db/2, generate/2]).

start_link() ->
  gen_server:start_link({local, ?MODULE}, ?MODULE, [], []).

init([]) ->
  rand:seed(default),
  {ok, []}.

duplicate(1, Original, Constructed) -> lists:append(Original, Constructed);
duplicate(N, Original, Constructed) -> duplicate(N-1, Original, lists:append(Original, Constructed)).

duplicate(List, N) -> duplicate(List, N, []).

padding(Size) when Size >= 4 -> (4 - (Size rem 4)) rem 4.

fill_padding_match([], Teams, Filled) -> {Filled, Teams};
fill_padding_match([padding | Left], [Picked | Teams], Filled) ->
  Duplicate = lists:member(Picked, Left) or lists:member(Picked, Filled),
  if Duplicate -> fill_padding_match([padding | Left], lists:append(Teams, [Picked]), Filled);
     true -> fill_padding_match(Left, Teams, [Picked | Filled])
  end;
fill_padding_match([Team | Left], Teams, Filled) ->
  fill_padding_match(Left, Teams, [Team | Filled]).

fill_padding_match(Match, Teams) ->
  fill_padding_match(Match, Teams, []).

fill_padding_schedule(Matches, Teams) ->
  TeamsRandom = [X || {_, X} <- lists:sort([{rand:uniform(), N} || N <- Teams])],
  {Schedule, _} = lists:mapfoldl(fun fill_padding_match/2, TeamsRandom, Matches),
  Schedule.

hash_teams_list(Teams) ->
  [ Y || <<X:4>> <= crypto:hash(sha, Teams), Y <- integer_to_list(X,16)].

generate(Process, {DatabaseFile, Division, Round, Seed, Matches, Teams}) -> gen_server:call(Process, {new_schedule, DatabaseFile, Division, Round, Seed, Matches, Teams}).
init_db(Process, DatabaseFile) -> gen_server:call(Process, {init_db, DatabaseFile}).

init_db(File) ->
  IsFile = filelib:is_regular(File),
  if IsFile -> file:delete(File);
     true -> true end,
  {ok, Database} = sqlite3:open(anonymous, [{file, File}]),
  ok = sqlite3:create_table(Database, teams, [{name, text, [not_null]},
                                              {inspection, blob}],
                            [{primary_key, [name]}]),
  ok = sqlite3:create_table(Database, matches, [{division, integer, [not_null]},
                                                {round, text, [not_null]},
                                                {size, integer, [not_null]},
                                                {teams, text, [not_null]},
                                                {number, integer, [not_null]},
                                                {seed, integer, [not_null]},
                                                {red_1, text, [not_null]},
                                                {red_2, text, [not_null]},
                                                {blue_1, text, [not_null]},
                                                {blue_2, text, [not_null]}],
                           [{primary_key, [division, round, size, teams, number, seed]},
                            {check, "round IN ('practice', 'qualification', 'elimination')"},
                            {foreign_key, {[blue_1], teams, [name], "ON DELETE RESTRICT"}},
                            {foreign_key, {[blue_2], teams, [name], "ON DELETE RESTRICT"}},
                            {foreign_key, {[red_1], teams, [name], "ON DELETE RESTRICT"}},
                            {foreign_key, {[red_2], teams, [name], "ON DELETE RESTRICT"}}]),
  ok = sqlite3:create_table(Database, match_scores, [{division, integer, [not_null]},
                                              {round, text, [not_null]},
                                              {size, integer, [not_null]},
                                              {teams, text, [not_null]},
                                              {number, integer, [not_null]},
                                              {seed, integer, [not_null]},
                                              {instance, integer, [not_null]},
                                              {score, blob, [not_null]}],
                           [{primary_key, [division, round, size, teams, number, seed, instance]},
                            {foreign_key, {[division, round, size, teams, number, seed], matches, [division, round, size, teams, number, seed], "ON DELETE CASCADE"}}]),
  ok = sqlite3:create_table(Database, match_states, [{division, integer, [not_null]},
                                              {round, text, [not_null]},
                                              {size, integer, [not_null]},
                                              {teams, text, [not_null]},
                                              {number, integer, [not_null]},
                                              {seed, integer, [not_null]},
                                              {time, integer, [not_null]},
                                              {state, blob, [not_null]}],
                            [{primary_key, [division, round, size, teams, number, seed, time]},
                             {foreign_key, {[division, round, size, teams, number, seed], matches, [division, round, size, teams, number, seed], "ON DELETE CASCADE"}}]),

  % TODO: finals
  % ok = sqlite3:create_table(Database, finals, [], [{primary_key, []}]),
  ok = sqlite3:create_table(Database, skills_scores, [{team, text, [not_null]},
                                                      {type, text, [not_null]},
                                                      {attempt, integer, [not_null]},
                                                      {score, blob, [not_null]}],
                            [{primary_key, [team, type, attempt]},
                             {foreign_key, {[team], teams, [name], "ON DELETE CASCADE"}}]),
  sqlite3:close(Database).

write(DatabaseFile, Division, Round, Seed, Matches, Teams) ->
  MatchList = create_schedule(Seed, Matches, Teams),
  Schedule = [[{division, Division},
               {round, atom_to_list(Round)},
               {size, Matches},
               {teams, hash_teams_list(Teams)},
               {number, N},
               {seed, Seed},
               {blue_1, B1},
               {blue_2, B2},
               {red_1, R1},
               {red_2, R2}] || {N, [B1, B2, R1, R2]} <- lists:enumerate(MatchList)],
  {ok, Database} = sqlite3:open(anonymous, [{file, DatabaseFile}]),
  [ok | Resp] = sqlite3:write_many(Database, matches, Schedule),
  ok = lists:last(Resp),
  ok = sqlite3:close(Database).


create_schedule(Seed, Matches, Teams) ->
  TeamsRepeated = duplicate(Matches, Teams),
  TeamsPadded = TeamsRepeated ++ lists:duplicate(padding(length(Teams)*Matches), padding),
  rand:seed(default, Seed),
  TeamsRandom = [X || {_, X} <- lists:sort([{rand:uniform(), N} || N <- TeamsPadded])],
  fill_padding_schedule(pick_matches(TeamsRandom), Teams).

pick_teams(Teams, Picked, 0, Safety) when Safety > 0 -> {Teams, Picked};
pick_teams([Team | Teams], Picked, N, Safety) when Safety > 0 ->
  Duplicate = lists:member(Team, Picked),
  if Duplicate, Team =/= padding -> pick_teams(lists:append(Teams, [Team]), Picked, N, Safety - 1);
     true -> pick_teams(Teams, [Team | Picked], N-1, Safety)
  end.

pick_teams(Teams, N) -> pick_teams(Teams, [], N, 1000).

pick_matches([], Matches) -> Matches;
pick_matches(Teams, Matches) ->
  {NewTeams, Match} = pick_teams(Teams, 4),
  pick_matches(NewTeams, [Match | Matches]).

pick_matches(Teams) -> pick_matches(Teams, []).

handle_cast(_, State) -> {noreply, State}.

handle_call({init_db, DatabaseFile}, _, State) ->
  {reply, init_db(DatabaseFile), State};
handle_call({new_schedule, DatabaseFile, Division, Round, Seed, Matches, Teams}, _, State) ->
  {reply, write(DatabaseFile, Division, Round, Seed, Matches, Teams), State}.
